package com.kob.backend.consumer;

// WebSocket用于前后端通信

import com.alibaba.fastjson2.JSONObject;
import com.kob.backend.config.RestTemplateConfig;
import com.kob.backend.consumer.utils.Game;
import com.kob.backend.consumer.utils.JwtAuthenticationUtil;
import com.kob.backend.mapper.RecordMapper;
import com.kob.backend.mapper.UserMapper;
import com.kob.backend.pojo.User;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.web.WebProperties;
import org.springframework.stereotype.Component;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;

import javax.websocket.*;
import javax.websocket.server.PathParam;
import javax.websocket.server.ServerEndpoint;
import javax.xml.stream.events.StartDocument;
import java.io.IOException;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArraySet;

@Component
@ServerEndpoint("/websocket/{token}")   // 注意不要以'/'结尾
public class WebSocketServer {
    /*
        存储所有链接:对所有 websocket 可见的全局变量,存储为 static 静态变量
        因为每个 websocket 实例都在一个独立的线程里,所以该公共变量应该是线程安全的
        使用线程安全的 Hash 表 ConcurrentHashMap<>(),
        将 userId 映射到 webSocketServer
    */
    public static final ConcurrentHashMap<Integer, WebSocketServer> users = new ConcurrentHashMap<>();
    // 用于和 MatchingSystem 进行通信
    private static RestTemplate restTemplate;
    // 在 WebSocketServer 中注入数据库的方法演示-> 使用 static 定义为独一份的变量
    private static UserMapper userMapper;
    // 注入 RecordMapper 用于调用实现游戏数据到数据库的存储
    public static RecordMapper recordMapper;
    // 后端向前端发送信息,首先需要创建一个 session
    private Session session = null;
    // 用户信息:定义一个成员变量
    private User user;
    private Game game = null;
    // addPlayer 添加用户到匹配池的 URL; removePlayer 从匹配池移除用户的 URL
    private final static String addPlayerUrl = "http://127.0.0.1:3001/player/add/";
    private final static String removePlayerUrl = "http://127.0.0.1:3001/player/remove/";

    @Autowired
    public void setRestTemplate(RestTemplate restTemplate) {
        // 注入 RestTemplate:作用是微服务之间发送请求
        WebSocketServer.restTemplate = restTemplate;
    }

    // 注入方法
    @Autowired
    public void setUserMapper(UserMapper userMapper) {
        // 静态变量 userMapper 访问需要使用类名 WebSocketServer 访问
        WebSocketServer.userMapper = userMapper;
    }

    @Autowired
    public void setRecordMapper(RecordMapper recordMapper) {
        WebSocketServer.recordMapper = recordMapper;
    }


    // @OnOpen 创建链接时自动触发这个函数
    @OnOpen
    public void onOpen(Session session, @PathParam("token") String token) throws IOException {
        // 建立链接时需要将 session 存下来
        this.session = session;
        // 成功建立连接时,输出 connected!
        System.out.println("backend connected!");
        // 将 userId 取出来,通过 userId 将 user 找出来
        // int userId = Integer.parseInt(token);
        int userId = JwtAuthenticationUtil.getUserId(token);
        this.user = userMapper.selectById(userId);

        // 如果 user 存在, 表示用户登录成功, 用户信息是存在的
        if (this.user != null) {
            // 将 user 存到 users HashMap里
            users.put(userId, this);

            // (测试)后台输出看用户信息
            // System.out.println(user);
        }
        // 否则,断开连接(这里需要抛出异常)
        else {
            this.session.close();
        }


    }

    @OnClose
    public void onClose() {
        // 关闭链接
        System.out.println("backend disconnected!");
        // 断开连接时,需要将 user 从 users 里面删掉
        if (this.user != null) {
            users.remove(this.user.getId());
        }
    }

    // 匹配逻辑函数
    public static void startGame(Integer aId, Integer bId) {
        User a = userMapper.selectById(aId);
        User b = userMapper.selectById(bId);

        // 匹配成功时,创建联机地图
        Game game = new Game(13, 14, 20, a.getId(), b.getId());
        game.createMap();   // 初始化地图
        // 当用户不为空时,才能执行下面的操作(防止有人退出游戏后用户已经成为空指针)
        if (users.get(a.getId()) != null)
            users.get(a.getId()).game = game;   // 将 game 赋给 a 玩家

        if (users.get(b.getId()) != null)
            users.get(b.getId()).game = game;

        game.start();   // 开启新线程,执行函数

        JSONObject respGameData = new JSONObject();
        respGameData.put("game_map", game.getG());
        respGameData.put("rows", game.getRows());
        respGameData.put("cols", game.getCols());
        respGameData.put("inner_walls_count", game.getInnerWallsCount());
        respGameData.put("a_id", game.getPlayerA().getId());
        respGameData.put("a_sx", game.getPlayerA().getSx());
        respGameData.put("a_sy", game.getPlayerA().getSy());
        respGameData.put("b_id", game.getPlayerB().getId());
        respGameData.put("b_sx", game.getPlayerB().getSx());
        respGameData.put("b_sy", game.getPlayerB().getSy());

        // 将 a 配对成功的消息传回客户端
        JSONObject respA = new JSONObject();
        respA.put("event", "start-matching");
        respA.put("opponent_username", b.getUsername());
        respA.put("opponent_photo", b.getPhoto());
        respA.put("game_data", respGameData);
        // 获取 a 的链接,并通过 sendMessage 将消息传给前端
        if (users.get(a.getId()) != null)
            users.get(a.getId()).sendMessage(respA.toJSONString());

        // 同理,将 b 的匹配成功信息传回给前端
        JSONObject respB = new JSONObject();
        respB.put("event", "start-matching");
        respB.put("opponent_username", a.getUsername());
        respB.put("opponent_photo", a.getPhoto());
        respB.put("game_data", respGameData);
        if (users.get(b.getId()) != null)
            users.get(b.getId()).sendMessage(respB.toJSONString());
    }

    // 开始匹配的逻辑部分:开始匹配时向 MatchingSystem 服务端发一个请求
    private void startMatching() {
        System.out.println("start matching");
        // 这里的参数列表需要与 MatchingSystem(微服务)->MatchingController->addPlayer 的参数对应
        MultiValueMap<String, String> data = new LinkedMultiValueMap<>();
        data.add("user_id", this.user.getId().toString());
        data.add("rating", this.user.getRating().toString());
        // postForObject(请求的URL,传出的数据,返回值的类型.class) -> Java 反射机制
        String res = restTemplate.postForObject(addPlayerUrl, data, String.class);
        System.out.println(this.user.getId().toString() + " " + this.user.getUsername() + " " + res);
    }

    // 取消匹配的逻辑部分:向 MatchingSystem 发送请求:从匹配池移除一个用户
    private void stopMatching() {
        System.out.println("stop matching");
        MultiValueMap<String, String> data = new LinkedMultiValueMap<>();
        data.add("user_id", this.user.getId().toString());
        String res = restTemplate.postForObject(removePlayerUrl, data, String.class);
        System.out.println(this.user.getId().toString() + " " + this.user.getUsername() + " " + res);
    }

    // direction 传入 move(移动) 方向参数
    private void move(int direction) {
        // 判断角色:如果是 A 角色,则将获取到的方向设置为 A 的 nextStep 方向
        // user.getId() 是获取当前链接的用户 id
        if (game.getPlayerA().getId().equals(user.getId())) {
            game.setNextStepA(direction);
        } else if (game.getPlayerB().getId().equals(user.getId())) {
            game.setNextStepB(direction);
        }
    }

    // @OnMessage 用于从前端接收请求: 一般 onMessage 当成路由使用,做为消息判断处理的中间部分
    @OnMessage
    public void onMessage(String message, Session session) {
        // 从 Client 接收消息
        System.out.println("receive message");
        // 解析从前端接收到的请求
        JSONObject data = JSONObject.parseObject(message);
        String event = data.getString("event");
        // System.out.println(event);

        // 匹配状态函数调用
        if ("start-matching".equals(event)) {
            startMatching();
        } else if ("stop-matching".equals(event)) {
            stopMatching();
        } else if ("move".equals(event)) {
            move(data.getInteger("direction"));
        }
    }

    @OnError
    public void onError(Session session, Throwable error) {
        error.printStackTrace();
    }

    // 从后端向前端发送信息
    public void sendMessage(String message) {
        // 异步通信过程,先加一个锁
        synchronized (this.session) {
            try {
                // 将 message 发送到前端
                this.session.getBasicRemote().sendText(message);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

}


